"""sqlamock: A SQL Database Mocking Library
==========================================

Usage Examples
--------------

1. Mocking tables using a dictionary:

    def test_default_interface(db_mock):
        with db_mock.from_dict({"table": [{"column": "other_column"}]}) as mocked_data:
            fetched_column = fetch_table_by_column("other_column")
            assert fetched_column.id == mocked_data["table"][0]["id"]

2. Mocking tables using ORM models:

    def test_using_orm_classes(db_mock):
        with db_mock.from_orm([MyModel(column="foo")]) as mocked_data:
            fetched_column = fetch_table_by_column("other_column")
            assert fetched_column.id == mocked_data[MyModel][0]["id"]

3. Loading mock data from a file:

    def test_loading_from_file(db_mock):
        with db_mock.from_file("mock_data/bulk_operations.json") as mocked_data:
            pass

Configuration
-------------

from typing import TYPE_CHECKING
from unittest.mock import patch

import pytest

from db import *  # noqa # load schemas
from models.base import BaseModel
from sqlamock.fixtures import *  # noqa

if TYPE_CHECKING:
    from sqlamock import MockConnectionProvider, Patches


@pytest.fixture(scope="session")
def db_mock_base_model():
    return BaseModel


@pytest.fixture(scope="session")
def db_mock_session(db_mock_patches: "Patches", db_mock_connection: "MockConnectionProvider"):
    db_mock_patches.add_patch(patch("app.models.connect.SessionLocal", db_mock_connection.get_session))



Example Test
------------

def test_model_create(db_mock: "DBMock", db_mock_connection: "MockConnectionProvider"):
    session = db_mock_connection.get_session()
    with db_mock.from_orm(
        [Model()]
    ) as mocked_data:
        repo = session.query(Model).first()
        assert repo.id == mocked_data[Model][0].id

"""

# Easier imports for typing or whatever :)
from .connection_provider import MockConnectionProvider
from .data_interface import MockDataInterface
from .db_mock import DBMock
from .patches import Patches

__all__ = [
    "DBMock",
    "MockConnectionProvider",
    "MockDataInterface",
    "Patches",
]



import tempfile
from contextlib import ExitStack
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from sqlalchemy import Engine

    from .connection_provider import ConnectionProvider


class Snapshot(ExitStack):
    """ContextManager that helps separate the scopes of database mocked data contexts.

    This class extends ExitStack to provide a mechanism for capturing and restoring
    the state of a database context. It is opened at the start of a db_mock context to
    create a snapshot of the pre-existing database state. When the db_mock context is exited,
    the snapshot is restored, effectively resetting anything that happened during the db_mock context.

    This is a critical part of teardown and maintaining deterministic testing state. This can only be used
    as a context manager within a db_mock context.

    Not meant for public use.

    Attributes:
           engine (Engine): The recyclable engine specific to the snapshot context.
           tmpfile_name (str): The name of the temporary file used to store the snapshot.
           connection_provider (MockConnectionProvider): The connection provider for the database.
    """

    if TYPE_CHECKING:
        engine: "Engine"
        tmpfile_name: str
        connection_provider: "ConnectionProvider"

    def __init__(self, connection_provider: "ConnectionProvider"):
        """Initialize a new Snapshot instance.

        Args:
            connection_provider (MockConnectionProvider): The connection provider for the database.
        """
        self.connection_provider = connection_provider
        super().__init__()

    def __enter__(self):
        """Enter the context manager, creating a snapshot of the current database state.

        This method creates a temporary file and dumps the current database state into it.

        Returns:
            Snapshot: The Snapshot instance.
        """
        tmpfile = tempfile.NamedTemporaryFile()
        self.tmpfile_name = self.enter_context(tmpfile).name
        with self.connection_provider.get_engine().connect() as conn:
            with open(self.tmpfile_name, "w") as f:
                for line in conn.connection.iterdump():
                    f.write(f"{line}\n")
        return super().__enter__()

    def __exit__(self, exc_type, exc_value, traceback):
        """Exit the context manager, restoring the database to the snapshotted state.

        This method resets the connection provider and restores the database state
        from the temporary file created in __enter__.

        Args:
            exc_type: The type of the exception that caused the context to be exited.
            exc_value: The instance of the exception that caused the context to be exited.
            traceback: A traceback object encoding the stack trace.
        """
        self.connection_provider.reset()
        with self.connection_provider.get_engine().connect() as conn:
            with open(self.tmpfile_name) as f:
                conn.connection.executescript(f.read())

        return super().__exit__(exc_type, exc_value, traceback)



import json
from contextlib import asynccontextmanager
from functools import cached_property
from typing import TYPE_CHECKING, Generic

from sqlalchemy import Index, Integer, inspect
from sqlalchemy.schema import CreateIndex, CreateTable

from sqlamock.patches import Patches

from .async_snapshot import AsyncSnapshot
from .data_interface import MockDataInterface
from .types import BaseType

if TYPE_CHECKING:
    from collections.abc import Iterable
    from contextlib import AbstractAsyncContextManager
    from pathlib import Path
    from typing import AsyncIterator

    from sqlalchemy import MetaData
    from sqlalchemy.ext.asyncio import AsyncEngine

    from .async_connection_provider import MockAsyncConnectionProvider
    from .data_interface import MockDataInterface


class AsyncDBMock(Generic[BaseType]):
    """A class that provides a mock interface for simulating interactions with
    multiple database tables in unit tests.

    This class validates the mocked data against the current state of the SQLAlchemy
    schemas, ensuring that the mock data aligns with the actual table and column
    definitions in the database models. The context managers returned by this class
    can be scoped to the fixture they are provided in, ensuring that each test
    maintains isolated and clean data. Different database data mocks can be applied
    at different scopes and layers. For example, you can provide default data for
    all tests, but within a test, apply additional data for a specific test function,
    or for managing multiple data contexts within the same test.

    Features:
    ---------
    - **Mock Data Across Tables**: Provides an interface to mock data for multiple
      tables, validated against the current SQLAlchemy schemas.
    - **Fixture Scoped Data**: The mocked data is scoped according to the fixture
      (e.g., function, module) it is provided in, ensuring that each test maintains
      isolated and clean data.
    - **No External Resources**: Supports unit testing without needing separately
      provisioned resources, such as a real database, allowing tests to run faster
      and in isolation.
    - **Foreign Key Support**: Allows defining foreign keys and creating rows across
      related tables, ensuring that relationships between tables are respected in
      the mock data.
    - **Isolated Mock Instances**: Each mock instance maintains its own state,
      allowing multiple test cases or fixtures to apply separate layers of data
      without conflict.
        - Example: You can have a default mock database state and apply additional
          test-specific data layers that are cleaned up when out of scope.
        - Limitation: Constraints across different mock instances will break. All
          data related to a test must be created within a single mock instance to
          remain isolated and valid.

    Methods:
    --------
    from_dict(data: dict[str, list[dict]]):
        Mocks multiple tables and their rows using a dictionary format, ensuring
        data consistency with SQLAlchemy schemas.

    from_orm(models: list[Base]):
        Mocks multiple database tables using SQLAlchemy ORM model instances,
        supporting relationships between tables and foreign keys.

    from_file(file_path: Path | str):
        Loads mock data for multiple tables from a JSON file, simulating tables
        with relationships, and bulk operations.
    """

    if TYPE_CHECKING:
        base: type[BaseType]
        connection_provider: "MockAsyncConnectionProvider"
        database_initialized: bool
        patches: Patches

    def __init__(
        self,
        base: "type[BaseType]",
        connection_provider: "MockAsyncConnectionProvider",
        patches: "Patches",
    ):
        self.base = base
        self.connection_provider = connection_provider
        self.database_initialized = False
        self.patches = patches

    @property
    def metadata(self) -> "MetaData":
        return self.base.metadata

    @cached_property
    def orm_classes(self) -> dict[str, "type[BaseType]"]:
        return {
            mapper.entity.__tablename__: mapper.entity
            for mapper in self.base.registry.mappers
        }

    def from_dict(
        self, data: dict[str, list[dict]]
    ) -> "AbstractAsyncContextManager[MockDataInterface]":
        """Mock multiple tables and their rows using a dictionary.

        Args:
        -----
        data (dict): Dictionary where the key is the table name and the value is
                     a list of rows (each row being a dictionary of column data).

        Returns:
        -------
        ContextManager[MockedDataInterface]: Mocked data interface containing
                                             created data by table and rows.
        """
        instances = [
            self.orm_classes[table_name](**row)
            for table_name, rows in data.items()
            for row in rows
        ]
        return self.from_orm(instances)

    def from_file(
        self, file_path: "Path | str"
    ) -> "AbstractAsyncContextManager[MockDataInterface]":
        """Load mock data for multiple tables from a JSON file and simulate
        relationships between tables.

        Args:
        -----
        file_path (str): Path to the JSON file containing mock data for multiple tables.

        Returns:
        -------
        ContextManager[MockedDataInterface]: Mocked data interface containing
                                             created data by table and rows.
        """
        with open(file_path) as f:
            data = json.load(f)
        return self.from_dict(data)

    @asynccontextmanager
    async def from_orm(
        self, instances: "Iterable[BaseType]"
    ) -> "AsyncIterator[MockDataInterface, None]":
        """Mock multiple database tables using SQLAlchemy ORM model instances.

        Args:
        -----
        instances (list): List of SQLAlchemy ORM model instances representing
                          rows in the database tables.

        Returns:
        -------
        Generator[MockedDataInterface, None, None]: A generator yielding a mocked
                                                    data interface containing
                                                    created data by table and rows.
        """
        with self.patches:
            await self.init_database()
            async with AsyncSnapshot(self.connection_provider):
                async with self.connection_provider.get_async_session() as session:
                    session.add_all(instances)
                    await session.commit()
                    for instance in instances:
                        await session.refresh(instance)

                db_mock_context: MockDataInterface = MockDataInterface(
                    instances=instances
                )
                yield db_mock_context

    async def init_database(self):
        if self.database_initialized:
            return

        tables_to_create = set()

        engine: "AsyncEngine" = self.connection_provider.get_async_engine()
        async with engine.connect() as conn:
            inspection = await conn.run_sync(inspect)

            for table in self.base.metadata.sorted_tables:
                # Check if the table exists
                table_exists = await conn.run_sync(
                    lambda sync_conn, t=table: inspection.has_table(
                        t.name, schema=t.schema
                    )
                )
                if not table_exists:
                    tables_to_create.add(table)

                pk_columns = [c for c in table.columns if c.primary_key]
                has_composite_pk = len(pk_columns) > 1

                # Address SQLite's limitation with Identity columns
                for column in table.columns:
                    if column.primary_key and column.type.python_type is int:
                        # only set autoincrement=True if not a composite PK
                        # sqlite does not support composite primary keys
                        column.autoincrement = not has_composite_pk
                        column.type = Integer()

            # Create tables asynchronously
            for table_to_create in tables_to_create:
                create_table_stmt = CreateTable(table_to_create)
                await conn.execute(create_table_stmt)

            # Create indexes asynchronously
            # CreateTable doesn't automatically create indexes, so we need to create them all manually
            for table in self.base.metadata.sorted_tables:
                existing_indexes = await conn.run_sync(
                    lambda sync_conn, t=table: inspection.get_indexes(
                        t.name, schema=t.schema
                    )
                )
                existing_index_names = {
                    idx_dict["name"] for idx_dict in existing_indexes
                }

                indexes_to_process = list(table.indexes)
                for index in indexes_to_process:
                    if index.name and index.name not in existing_index_names:
                        try:
                            index_to_create = index
                            if hasattr(index, "dialect_options"):
                                postgresql_opts = index.dialect_options.get(
                                    "postgresql", {}
                                )
                                postgresql_where = postgresql_opts.get("where")
                                if postgresql_where is not None:
                                    index_to_create = Index(
                                        index.name,
                                        *[col for col in index.columns],
                                        unique=index.unique,
                                        sqlite_where=postgresql_where,
                                    )
                            create_index_stmt = CreateIndex(index_to_create)
                            await conn.execute(create_index_stmt)
                        except Exception:
                            pass

        self.database_initialized = True



import asyncio
from typing import TYPE_CHECKING

from .snapshot import Snapshot

if TYPE_CHECKING:
    from sqlalchemy import Engine

    from .async_connection_provider import MockAsyncConnectionProvider


class AsyncSnapshot(Snapshot):
    if TYPE_CHECKING:
        engine: "Engine"
        tmpfile_name: str
        connection_provider: "MockAsyncConnectionProvider"

    def __init__(self, connection_provider: "MockAsyncConnectionProvider"):
        """Initialize a new Snapshot instance.

        Args:
            connection_provider (MockConnectionProvider): The connection provider for the database.
        """
        self.connection_provider = connection_provider
        super().__init__(connection_provider)

    async def __aenter__(self):
        """Enter the context manager, creating a snapshot of the current database state.

        This method creates a temporary file and dumps the current database state into it.

        Returns:
            Snapshot: The Snapshot instance.
        """
        return await asyncio.to_thread(self.__enter__)

    async def __aexit__(self, exc_type, exc_value, traceback):
        """Exit the context manager, restoring the database to the snapshotted state.

        This method resets the connection provider and restores the database state
        from the temporary file created in __enter__.

        Args:
            exc_type: The type of the exception that caused the context to be exited.
            exc_value: The instance of the exception that caused the context to be exited.
            traceback: A traceback object encoding the stack trace.
        """
        await self.connection_provider.async_reset()
        return await asyncio.to_thread(self.__exit__, exc_type, exc_value, traceback)



import asyncio
from functools import lru_cache
from typing import TYPE_CHECKING

from sqlalchemy.ext.asyncio import AsyncEngine, AsyncSession, create_async_engine

from sqlamock.connection_provider import MockConnectionProvider


class MockAsyncConnectionProvider(MockConnectionProvider):
    """A class that provides mock database connections for patching purposes.

    This class creates and manages SQLite in-memory(file) databases, which can be used
    as a lightweight replacement for actual database connections in tests.

    Attributes:
        engine_kwargs (dict): Additional keyword arguments to pass to create_engine.
    """

    if TYPE_CHECKING:
        engine_kwargs: dict

    def __init__(self, engine_kwargs: dict | None = None):
        """Initialize a new MockAsyncConnectionProvider instance.

        Args:
            engine_kwargs (dict | None): Additional keyword arguments to pass to create_async_engine.
                                         If None, an empty dict will be used.
        """
        self.engine_kwargs = engine_kwargs or {}

    @lru_cache  # noqa: B019
    def get_async_engine(self) -> AsyncEngine:
        """Get or create a SQLAlchemy async engine instance.

        Returns:
            AsyncEngine: A SQLAlchemy async engine instance.
        """
        engine = self.get_engine()
        return create_async_engine(
            engine.url.set(drivername="sqlite+aiosqlite"),
            **self.engine_kwargs,
        )

    def get_async_session(self) -> AsyncSession:
        """Create a new SQLAlchemy async session.

        Returns:
            AsyncSession: A new SQLAlchemy async session instance.
        """
        return AsyncSession(bind=self.get_async_engine())

    async def async_reset(self):
        """Reset the connection provider.

        This method disposes of the current engine and clears the engine cache,
        ensuring that a new engine will be created on the next call to get_engine().

        This is used in conjunction with the Snapshot context manager to reset the
        database state between db_mock contexts (especially nested ones).
        """
        await asyncio.to_thread(self.reset)
        await self.get_async_engine().dispose()
        self.get_async_engine.cache_clear()



import tempfile
from functools import lru_cache
from pathlib import Path
from typing import TYPE_CHECKING

from sqlalchemy import Engine, create_engine
from sqlalchemy.orm import Session


class MockConnectionProvider:
    """A class that provides mock database connections for patching purposes.

    This class creates and manages SQLite in-memory(file) databases, which can be used
    as a lightweight replacement for actual database connections in tests.

    Attributes:
        engine_kwargs (dict): Additional keyword arguments to pass to create_engine.
    """

    if TYPE_CHECKING:
        engine_kwargs: dict

    def __init__(self, engine_kwargs: dict | None = None):
        """Initialize a new MockConnectionProvider instance.

        Args:
            engine_kwargs (dict | None): Additional keyword arguments to pass to create_engine.
                                         If None, an empty dict will be used.
        """
        self.engine_kwargs = engine_kwargs or {}

    @lru_cache  # noqa: B019
    def get_engine(self) -> Engine:
        """Get or create a SQLAlchemy engine instance.

        This method creates a new SQLite in-memory database engine. The result is
        cached, so subsequent calls will return the same engine instance.

        Returns:
            Engine: A SQLAlchemy engine instance.
        """
        with tempfile.NamedTemporaryFile() as tmpfile:
            return create_engine(f"sqlite:///{tmpfile.name}", **self.engine_kwargs)

    def get_session(self) -> Session:
        """Create a new SQLAlchemy session.

        This method creates a new session bound to the engine returned by get_engine().

        Returns:
            Session: A new SQLAlchemy session instance.
        """
        return Session(bind=self.get_engine())

    def reset(self):
        """Reset the connection provider.

        This method disposes of the current engine and clears the engine cache,
        ensuring that a new engine will be created on the next call to get_engine().

        This is used in conjunction with the Snapshot context manager to reset the
        database state between db_mock contexts (especially nested ones).
        """
        self.get_engine().dispose()
        self.get_engine.cache_clear()



from contextlib import ExitStack
from threading import Lock
from typing import TYPE_CHECKING
from unittest.mock import patch

from sqlalchemy.exc import IntegrityError
from sqlalchemy.sql.sqltypes import Enum

if TYPE_CHECKING:
    from contextlib import AbstractContextManager
    from typing import Self


def mock_enum():
    original = Enum._object_value_for_elem

    def _(*args, **kwargs):
        try:
            return original(*args, **kwargs)
        except LookupError as e:
            raise IntegrityError(
                (
                    "(psycopg.errors.InvalidTextRepresentation) "
                    f"invalid input value for enum '{kwargs.get('column_name', 'unknown_enum')}': '{kwargs.get('value', 'unknown_value')}'\n"
                    "DETAIL:  The provided value is not an existing member of the enum set."
                ),
                params=None,
                orig=e,
            ) from e

    return patch.object(Enum, "_object_value_for_elem", _)


class Patches(ExitStack):
    """Manages SQLAlchemy patches and other db related patches for testing within a db_mock context.

    It can handle nested contexts safely using an internal lock and counter, applying patches only once when the first context is entered,
    and restoring them when the final context is exited.

    Attributes:
        patches (list[AbstractContextManager]): A list of additional patches added by the user.
        counter (int): Tracks the number of active contexts to control the application and restoration of patches.
    """

    if TYPE_CHECKING:
        _lock: Lock
        patches: list[AbstractContextManager]
        counter: int

    def __init__(self, *args, **kwargs):
        """Initializes the Patches context manager.

        This sets up the necessary lock, initializes the patch list and counter, and
        inherits from `ExitStack` for efficient context management.
        """
        super().__init__(*args, **kwargs)
        self._lock = Lock()
        self.patches = [mock_enum()]
        self.counter = 0

    def add_patch(self, patch: "AbstractContextManager"):
        """Adds an additional patch to be applied within the db_mock context.

        This method allows adding extra patches, which will be applied along with
        `mock_enum` when the context is entered. This is useful for any custom patches
        needed in conjunction with the main SQLAlchemy patches. This is defined at the session
        scope.

        Args:
            patch (AbstractContextManager): The patch context manager to be added.
        """
        self.patches.append(patch)

    def __enter__(self) -> "Self":
        """db_mock manages opening and closing Patches context, to apply the patches alongside the db_mock context.

        If this is the first context being entered (counter is 0), it applies all patches. The counter increments with each
        nested context to count the number of active contexts.

        Returns:
            Patches: The current instance of Patches, with patches applied.
        """
        with self._lock:
            if self.counter == 0:
                for patch in self.patches:
                    self.enter_context(patch)

            self.counter += 1

        return super().__enter__()

    def __exit__(self, *args, **kwargs) -> None:
        """Counts down the number of active contexts, restoring patches when the last nested context exits."""
        with self._lock:
            self.counter -= 1
            if self.counter == 0:
                super().__exit__(*args, **kwargs)



from collections import defaultdict
from typing import TYPE_CHECKING, Generic

from .types import BaseType

if TYPE_CHECKING:
    from collections.abc import Iterable


class MockDataInterface(Generic[BaseType]):
    """A class that provides an interface for accessing mocked database data.

    This class organizes and provides access to mocked database instances,
    allowing retrieval by both ORM class and table name for QOL purposes.

    This is returned by the db_mock entered context. It is primarily useful for retrieving
    dynamically determined column values like primary keys for further assertions.

    Attributes:
        data_registry (dict): A dictionary mapping ORM classes to lists of their instances.
        table_name_mapping (dict): A dictionary mapping table names to their corresponding ORM classes.
    """

    if TYPE_CHECKING:
        data_registry: dict[type[BaseType], list[BaseType]]
        table_name_mapping: dict[str, type[BaseType]]

    def __init__(self, instances: "Iterable[BaseType]"):
        """Initialize the MockDataInterface with a list of ORM instances.

        This method populates the data_registry and table_name_mapping
        based on the provided instances.

        Args:
            instances (list[BaseType]): A list of SQLAlchemy ORM instances.
        """
        self.data_registry = defaultdict(list)
        self.table_name_mapping = {}

        for instance in instances:
            self.data_registry[type(instance)].append(instance)

        for key in self.data_registry.keys():
            self.table_name_mapping[key.__tablename__] = key

    def __getitem__(self, key: "type[BaseType] | str") -> list["BaseType"]:
        """Retrieve mocked data instances by ORM class or table name.

        This method allows accessing mocked data using either the ORM class
        or the table name as a key.

        Args:
            key (type[BaseType] | str): The ORM class or table name to retrieve data for.

        Returns:
            list[BaseType]: A list of mocked data instances for the specified key.

        Raises:
            KeyError: If the provided key is not found in the data registry or table name mapping.
        """
        if isinstance(key, str):
            try:
                return self.data_registry[self.table_name_mapping[key]]
            except KeyError as e:
                raise KeyError(f"Table name {key} not found") from e

        return self.data_registry[key]



from typing import TYPE_CHECKING

import pytest

from sqlamock.patches import Patches

from .async_connection_provider import MockAsyncConnectionProvider
from .async_db_mock import AsyncDBMock

if TYPE_CHECKING:
    from .types import BaseType


@pytest.fixture(scope="session")
def db_mock_async_connection() -> "MockAsyncConnectionProvider":
    return MockAsyncConnectionProvider()


@pytest.fixture(scope="session")
def db_mock_async(
    db_mock_async_connection: "MockAsyncConnectionProvider",
    db_mock_base_model: "type[BaseType]",
    db_mock_patches: "Patches",
) -> "AsyncDBMock":
    return AsyncDBMock(db_mock_base_model, db_mock_async_connection, db_mock_patches)



from typing import TYPE_CHECKING

import pytest
from sqlalchemy.orm import declarative_base

from sqlamock.patches import Patches

from .connection_provider import MockConnectionProvider
from .db_mock import DBMock

if TYPE_CHECKING:
    from .connection_provider import ConnectionProvider
    from .types import BaseType


@pytest.fixture(scope="session")
def db_mock_base_model() -> "type[BaseType]":
    """Fixture that is used as an interface to provide the base SQLAlchemy declarative model used
    in schemas to be tested.

    This also requires that your schemas are imported and registered to the base model at some point
    before db_mock context.

    Returns:
        type[SQLAlchemyORMProtocol]: SQLAlchemy declarative model
    """
    return declarative_base()


@pytest.fixture(scope="session")
def db_mock_connection() -> "ConnectionProvider":
    """Fixture that provides a mock database connection for testing.

    This fixture sets up a mock database connection using MockConnectionProvider.
    It is expected that this fixture is used in another fixture that patches the session or engine provider
    in the code to be tested.

    Note: the engine is recycled, so if the code caches an engine, this interface cannot be used to patch that engine
    Instead the session provider should be patched. Surely, the session is not cached...

    Yields:
        MockConnectionProvider: An instance of MockConnectionProvider for use in patching.
    """
    return MockConnectionProvider()


@pytest.fixture(scope="session")
def db_mock_patches() -> "Patches":
    return Patches()


@pytest.fixture(scope="session")
def db_mock(
    db_mock_connection: "ConnectionProvider",
    db_mock_base_model: "type[BaseType]",
    db_mock_patches: "Patches",
) -> "DBMock":
    """Fixture that provides the main DBMock interface for database data mocking in tests.

    This fixture creates and returns a DBMock instance, which can be used to
    mock database operations in tests. Please see db_mock.py::DBMock for more information.

    Args:
        db_mock_connection (MockConnectionProvider): The mock connection provider.
        db_mock_base_model (type[SQLAlchemyORMProtocol]): The base SQLAlchemy model.

    Returns:
        DBMock: An instance of DBMock for use in tests.
    """
    return DBMock(db_mock_base_model, db_mock_connection, db_mock_patches)
